#include "App.h"

#include <cstdlib>
#include <iostream>
#include <typeinfo>
#include <X11/Xutil.h>


////////////////////////////////////////////////////////////////////////////////
App::App(const std::string & appname)
   {
   XSetErrorHandler(&App::x_error_handler);

   if ((dpy = XOpenDisplay(NULL)) == NULL) 
      {
      std::cerr << "Unable to connect to display " << XDisplayName(NULL) << std::endl;
      exit(1);
      }
   std::cout << "Display name     : " << XDisplayName(NULL) << std::endl
	     << "Number of screens: " << ScreenCount(dpy) << std::endl
	     << "Default screen   : " << DefaultScreen(dpy) << std::endl
	     << "Protocol Version : " << ProtocolVersion(dpy) << std::endl
	     << "Protocol Revision: " << ProtocolRevision(dpy) << std::endl
	     << "Server Vendor    : " << ServerVendor(dpy) << std::endl
	     << "Vendor Release   : " << VendorRelease(dpy) << std::endl
	     << std::endl
	     << "Display width    : " << this->display_width() << std::endl
	     << "Display height   : " << this->display_height() << std::endl
	 ;

   App::BitMap * icon = new App::BitMap("xlib_App.xbm");

   // Create the main window
   create_window("main", 
		 0, 0 , 
		 this->display_width()/2, this->display_height()/2, 
		 icon);

   // Get a reference to our WindowData structure.  We need the Window handle from it for a few things.
   App::WindowData * wd = windows["main"];




   // We need to set the window manager hints, size hints, window name, and icon name

   // Even though we are not passing any actual size_hints, the Xlib programming manual says to pass the flags below anyway.
   XSizeHints *size_hints;
   if (!(size_hints = XAllocSizeHints())) 
      {
      std::cerr << "XAllocSizeHints: memory allocation failure" << std::endl;
      exit(1);
      }
   size_hints->flags = PPosition | PSize;

   // Get X Properties for the window and icon names.
   XTextProperty windowName, iconName;
   const char * ptr = appname.c_str();
   if (XStringListToTextProperty(const_cast<char **>(&ptr), 1, &windowName) == 0) 
      {
      std::cerr << "XStringListToTextProperty: structure allocation for windowName failed." << std::endl;;
      exit(1);
   }
   ptr = icon->name.c_str();
   if (XStringListToTextProperty(const_cast<char **>(&ptr), 1, &iconName) == 0) 
      {
      std::cerr << "XStringListToTextProperty: structure allocation for iconName failed." << std::endl;;
      exit(1);
   }

   XWMHints *wm_hints;
   if (!(wm_hints = XAllocWMHints()))
      {
      std::cerr << "XAllocWMHints: memory allocation failure" << std::endl;
      exit(0);
      }
   wm_hints->initial_state = NormalState;
   wm_hints->input = True;
   wm_hints->icon_pixmap = icon->pixmap;
   wm_hints->flags = StateHint | IconPixmapHint | InputHint;

   XSetWMProperties(dpy, wd->win, &windowName, &iconName,
    		    NULL, 0, size_hints, wm_hints,
		    NULL);



   // Select the event types we want to receive.
   //Other interesting events include KeyReleaseMask and ButtonReleaseMask
   event_mask = 
	 ExposureMask | StructureNotifyMask |
	 KeyPressMask | KeyReleaseMask | 
	 FocusChangeMask |
	 PointerMotionMask |
	 EnterWindowMask | LeaveWindowMask |
	 ButtonPressMask | ButtonReleaseMask ;
   XSelectInput(dpy, wd->win, event_mask);



   // Make sure we get delete events from the window manager.
   // "wm_delete_window" is the Atom which corresponds to the delete
   //  window message sent by the window manager. 

   wm_delete_window = XInternAtom (dpy, "WM_DELETE_WINDOW", False);
   XSetWMProtocols(dpy, wd->win, &wm_delete_window, 1);


   // This give BadMatch and I don't know why.
   // XSetWindowAttributes setwinattr;
   // setwinattr.background_pixmap = bg_pixmap->pixmap;
   // XChangeWindowAttributes(dpy, wd->win, CWBackPixmap, &setwinattr);

   // Also BadMatch
   // XSetWindowBackgroundPixmap(dpy, wd->win, icon->pixmap);


   //   XSetWindowBackgroundPixmap(dpy, wd->win, None);


   // Map the window.  Remember: this does not display the window immediately.
   // The request is queued until events are read, or we call XFlush or XSync.
   XMapWindow(dpy, wd->win);
   }

////////////////////////////////////////////////////////////////////////////////
App::~App()
   {
   // Ugh.  I really wish I could figure out why I can't make a map of string 
   // to WindowData instead of having ton use WindowData *
   // I'd like to eliminate this next loop.
   for( WindowMap::iterator i = windows.begin(); i != windows.end(); ++i)
      {
      App::WindowData * wp = (*i).second;
      windows.erase(i);
      delete wp;
      }

   if (dpy)
      {
      XCloseDisplay(dpy);
      }
   }

////////////////////////////////////////////////////////////////////////////////
int App::x_error_handler(Display * dpy, XErrorEvent *err)
   {
   char err_msg[1024];
   err_msg[0] = 0;
   XGetErrorText(dpy, err->error_code, err_msg, 1024);
   std::cerr << "X11 Error"<< std::endl
	     << "    display     : " << DisplayString(dpy) << std::endl
	     << "    serial      : " << err->serial << std::endl
	     << "    error_code  : " << (int) err->error_code << std::endl
	     << "    request_code: " << (int) err->request_code << std::endl
	     << "    minor_code  : " << (int) err->minor_code << std::endl
	     << err_msg << std::endl;
   return 0;
   }

////////////////////////////////////////////////////////////////////////////////
void App::create_window(const std::string & win_name,
			int x, int y,
			unsigned int width, unsigned int height,
			App::BitMap * icon)
   {
   Window win;
   int border_width = 4;           /* Border four pixels wide */

   int screen_num = DefaultScreen(dpy);

   /* Create opaque window */
   win = XCreateSimpleWindow(dpy, 
			     RootWindow(dpy, screen_num),
			     x, y, 
			     width, height, 
			     border_width, 
			     None, None);
   //			     BlackPixel(dpy, screen_num),
   //			     WhitePixel(dpy, screen_num));

   icon->make_pixmap(dpy, win);

   // Add it to the window list
   App::WindowData * win_data = new App::WindowData(win_name, win, icon);
   windows.insert(win_pair(win_name, win_data));
   }

////////////////////////////////////////////////////////////////////////////////
void App::run(void)
   {
   XEvent e;
   KeySym k;
   char key_buffer[8];
   key_buffer[7] = 0;
   XComposeStatus compose_status;

   int done = 0;

   while (!done)
      {
      while (XCheckIfEvent(dpy, &e, 
      			   // *sniff*  My very first lambda function in C++
      			   [&](Display * d, XEvent * e, XPointer a)->Bool { return True; }, 
      			   NULL))
	 {
	 switch (e.type)
	    {

	    ////////////////////////
	    // General window events
	    case Expose:
	       // We'll only redraw the entire window at a time, so unless this is 
	       // the last contiguous expose, don't draw the window.
	       if (e.xexpose.count != 0)
		  break;
	       // Draw the window contents here
	       break;
	    case ClientMessage:
	       if ((Atom)e.xclient.data.l[0] == wm_delete_window)
		  done = 1;
	       break;

	    ///////////////////////
	    // Mouse events
	    case MotionNotify:
	       std::cout << "Mouse at (" << e.xmotion.x << "," << e.xmotion.y << ")" << std::endl;
	       break;
	    case ButtonPress:
	       std::cout << "You pressed button " << e.xbutton.button << 
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;
	    case ButtonRelease:
	       std::cout << "You released button " << e.xbutton.button <<
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;
	    case EnterNotify:
	       std::cout << "Mouse has entered the window" <<
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;
	    case LeaveNotify:
	       std::cout << "Mouse has left the window" << 
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;

	    //////////////////
	    // Keyboard Events
	    case FocusIn:
	       // Interesting note: When we regain focus due to a mouse click in the window, 
	       // we first get a LeaveNotify, then the FocusIn, then an EnterNotify, and then
	       // the ButtonPress and ButtonRelease.  So in a real app we might want to have
	       // the FocusIn handler read ahead and handle (discard) the next ButtonPress,
	       // and ButtonRelease so that the click does not have a real effect on the application.
	       std::cout << "Got keyboard focus" << std::endl;
	       break;
	    case FocusOut:
	       std::cout << "Lost keyboard focus" << std::endl;
	       break;
	    case KeyPress:
	       // Doing the lookup like this, using XLookupString, means I get the modified
	       // character back.  Thus type shift+a gives me XK_A instead of XK_a
	       // If I used XKeySym I'd instead get a sequence such as this:
	       // XK_Shift_L followed by XK_a
	       //
	       // In the end, that might be a more useful way to go, but for now I'll keep
	       // things simple.
	       XLookupString(&(e.xkey), key_buffer, 7, &k, &compose_status);
	       switch (k)
		  {
		  case XK_Escape:
		  case XK_q:
		     std::cout << "Goodbye!" << std::endl;
		     done = 1;
		     break;
		  default:
		     std::cout << "You typed " << key_buffer << std::endl;
		     break;
		  }
	    case MappingNotify:
	       // While it is uncommon for the user to change the keyboard mapping while
	       // a program is running, handling this case is so trivial that there is 
	       // simply no excuse to leave it out.
	       XRefreshKeyboardMapping((XMappingEvent *)&e);
	       break;

	    default:
	       break;
	    }
	 }

      // Do idle time things here

      }
   }

////////////////////////////////////////////////////////////////////////////////
std::string App::button_state_string(unsigned int state) const
   {
   std::string str = "";
   if (state & Button1Mask) { str.append("Button1 "); }
   if (state & Button2Mask) { str.append("Button2 "); }
   if (state & Button3Mask) { str.append("Button3 "); }
   if (state & Button4Mask) { str.append("Button4 "); }
   if (state & Button5Mask) { str.append("Button5 "); }
   if (state & ShiftMask)   { str.append("Shift "); }
   if (state & LockMask)    { str.append("Lock "); }
   if (state & ControlMask) { str.append("Control "); }
   if (state & Mod1Mask)    { str.append("Mod1 "); }
   if (state & Mod2Mask)    { str.append("Mod2 "); }
   if (state & Mod3Mask)    { str.append("Mod3 "); }
   if (state & Mod4Mask)    { str.append("Mod4 "); }
   if (state & Mod5Mask)    { str.append("Mod5 "); }
   return str;
   }

////////////////////////////////////////////////////////////////////////////////
// App::BitMap
////////////////////////////////////////////////////////////////////////////////
App::BitMap::BitMap(const std::string & file)
   {
   read_from_file(file);
   }

////////////////////////////////////////////////////////////////////////////////
App::BitMap::~BitMap() 
   {
   if (data) { XFree(data); data = NULL; }
   }

////////////////////////////////////////////////////////////////////////////////
void App::BitMap::make_pixmap(Display * dpy, Drawable d)
   {
   pixmap = XCreateBitmapFromData(dpy, d, (char *) data, width, height);
   }

////////////////////////////////////////////////////////////////////////////////
int App::BitMap::read_from_file(const std::string & file)
   {
   if (BitmapSuccess != XReadBitmapFileData(file.c_str(), &width, &height, &data, NULL, NULL))
      {
      return 0;
      }

   name = file;
   size_t p = name.rfind("/");
   if (p != std::string::npos)
      {
      name.replace(0, p+1, "");
      }
   return 1;
   }

